import pickle
from abc import ABC

DEFAULT_DATA_DIR = "data"
DEFAULT_TRAINER_FILEPATH = DEFAULT_DATA_DIR + "/" + "trainer.pkl"

class Trainer(ABC):
    @staticmethod
    def load(datadir=DEFAULT_DATA_DIR):
        with open(datadir + "/trainer.pkl", "rb") as fin:
            return pickle.load(fin)

    @staticmethod
    def resume(datadir=DEFAULT_DATA_DIR, render: bool = False, verbosity:int=1):
        trainer = Trainer.load(datadir)
        return trainer.main(render=render, verbosity=verbosity)

    @staticmethod
    def eval(datadir=DEFAULT_DATA_DIR, render: bool = False, verbosity:int=1):
        trainer = Trainer.load(datadir)
        return trainer.main(eval=True, render=render, verbosity=verbosity)



from abc import abstractmethod
from typing import Set

from flloat.semantics.pl import PLInterpretation
from pythomata.base.Simulator import DFASimulator, Simulator
from pythomata.base.Symbol import Symbol
from pythomata.base.utils import Sink

from rltg.logic.CompleteRewardAutomaton import CompleteRewardAutomaton

class RewardSimulator(Simulator):

    def potential_function(self, q, is_terminal_state=False):
        raise NotImplementedError

    @abstractmethod
    def is_failed(self) -> bool:
        raise NotImplementedError


class RewardAutomatonSimulator(DFASimulator, RewardSimulator):
    def __init__(self, dfa:CompleteRewardAutomaton):
        super().__init__(dfa)
        self.visited_states = {self.cur_state}

    def reset(self):
        super().reset()
        self.visited_states = {self.cur_state}

    def make_transition(self, s:Set[Symbol], **kwargs):
        i = PLInterpretation(s)
        super().make_transition(i)
        self.visited_states.add(self.cur_state)
        return self.cur_state

    def potential_function(self, q, is_terminal_state=False):
        return self.dfa.potential_function(q, is_terminal_state=is_terminal_state)

    def is_failed(self):
        return self.id2state[self.cur_state] in self.dfa.failure_states

    def get_current_state(self):
        return self.cur_state



import logging
from copy import copy


class mydefaultdict(dict):
    def __init__(self, x):
        super().__init__()
        self._default = x

    def __getitem__(self, key):
        if key in self:
            return dict.__getitem__(self, key)
        else:
            self[key] = copy(self._default)
            return self[key]

def logger_from_verbosity(verbosity):
    verbosity = 0 if verbosity < 0 else 2 if verbosity > 2 else verbosity
    logging.getLogger().setLevel(logging.WARNING if verbosity==0 else logging.INFO if verbosity==1 else logging.DEBUG)

class AgentObservation(object):
    def __init__(self, state, action, reward, state2):
        self.state = state
        self.action = action
        self.reward = reward
        self.state2 = state2

    def unpack(self):
        return self.state, self.action, self.reward, self.state2


def compute_levels(dfa, property_states):
    assert property_states.issubset(dfa.states)
    level = 0
    state2level = {final_state: level for final_state in property_states}

    z_current = set()
    z_next = set(property_states)
    while z_current != z_next:
        level += 1
        z_current = z_next
        z_next = copy(z_current)
        for s in dfa.states:
            if s in z_current:
                continue
            for a in dfa.transition_function.get(s, []):
                next_state = dfa.transition_function[s][a]
                if next_state in z_current:
                    z_next.add(s)
                    state2level[s] = level

    z_current = z_next

    max_level = level - 1

    # levels for failure state (i.e. that cannot reach a final state)
    failure_states = set()
    for s in filter(lambda x: x not in z_current, dfa.states):
        state2level[s] = level
        failure_states.add(s)

    return state2level, max_level, failure_states


def _potential_function(q, initial_state, reachability_levels, reward, is_terminal_state=False):
    if is_terminal_state:
        return 0
    else:
        # p = 1/(self.reachability_levels[q]) * self.reward
        # if q == initial_state and reachability_levels[initial_state]==0:
        #     return reward
        initial_state_level = reachability_levels[initial_state]
        p = initial_state_level - reachability_levels[q]
        p = p/initial_state_level if initial_state_level!=0 else p
        p *= reward
    return p



import logging

from rltg.agents.TGAgent import TGAgent
from rltg.trainers.GenericTrainer import GenericTrainer
from rltg.utils.GoalEnvWrapper import GoalEnvWrapper
from rltg.utils.StatsManager import StatsManager
from rltg.utils.StoppingCondition import GoalPercentage, CheckAutomataInFinalState


class TGTrainer(GenericTrainer):
    """Temporal Goal Trainer"""

    def __init__(self, env:GoalEnvWrapper, agent:TGAgent=None, n_episodes=1000,
                 data_dir="data",
                 stop_conditions=(GoalPercentage(10, 1.0), ),
                 ):
        super().__init__(env, agent, n_episodes, data_dir, stop_conditions)
        self.stop_conditions.append(CheckAutomataInFinalState())

    def train_loop(self, render:bool=False):
        env = self.env
        agent = self.agent

        stop_condition = False
        info = {"goal": False}

        state = env.reset()
        action = agent.start(state)
        obs = None
        if render: env.render()

        while not stop_condition:
            state2, reward, done, info = env.step(action)
            if render: env.render()

            old_automata_state = agent.get_automata_state()
            new_automata_state = agent.sync(action, state2)

            stop_condition = self.check_episode_stop_conditions(done, info.get("goal", False))
            obs = agent.observe(state, action, reward, state2,
                                old_automata_state=old_automata_state,
                                new_automata_state=new_automata_state,
                                is_terminal_state=stop_condition)

            if stop_condition:
                break

            action = agent.step(obs)
            agent.update()
            state = state2

        agent.end(obs)
        steps, tot_reward = agent.brain.episode_iteration, agent.brain.total_reward
        return steps, tot_reward, self.is_goal(info)


    def check_stop_conditions(self, stats:StatsManager, eval=False, *args, **kwargs):
        return not eval and all([s.check_condition(stats_manager=stats, temporal_evaluators=self.agent.temporal_evaluators)
                    for s in self.stop_conditions])

    def check_episode_stop_conditions(self, done, goal, *args, **kwargs):
        temp_evals = self.agent.temporal_evaluators
        any_te_failed = any(t.is_failed() for t in temp_evals)
        all_te_true = all(t.is_true() for t in temp_evals) if len(temp_evals) > 0 else False

        if any_te_failed:
            logging.debug("some automaton in failure state: end episode")

        return done or any_te_failed or (goal and all_te_true)

    def is_goal(self, info, *args, **kwargs):
        return super().is_goal(info) and all(t.is_true() for t in self.agent.temporal_evaluators)



import logging
import os
import pickle
import re
from typing import List

from gym.spaces import Tuple

from rltg.agents.Agent import Agent
from rltg.agents.brains.Brain import Brain
from rltg.agents.feature_extraction import RobotFeatureExtractor
from rltg.agents.temporal_evaluator.TemporalEvaluator import TemporalEvaluator
from rltg.utils.RewardShaping import AutomatonRewardShaping, DynamicAutomatonRewardShaping


class TGAgent(Agent):
    """Temporal Goal agent"""

    def __init__(self,
                 sensors: RobotFeatureExtractor,
                 brain:Brain,
                 temporal_evaluators: List[TemporalEvaluator],
                 reward_shaping: bool = True):
        assert len(temporal_evaluators)>=1

        if reward_shaping:
            reward_shapers = tuple([AutomatonRewardShaping(te.simulator) if not te.on_the_fly else
                                    DynamicAutomatonRewardShaping(te.simulator) for te in temporal_evaluators])
        else:
            reward_shapers = tuple()
        super().__init__(sensors, brain, reward_shapers=reward_shapers)
        self.temporal_evaluators = temporal_evaluators

        # compute the feature space. It is the cartesian product between
        # the robot feature space output and the automata state space
        # sensors.output_space is expected to be a Tuple.
        robot_feature_space = sensors.output_space
        automata_state_spaces = [temp_eval.get_state_space() for temp_eval in temporal_evaluators]

        # total feature space = (robot feature space, automata 1 state space, automata 2 state space, ... )
        feature_space = Tuple(tuple([robot_feature_space.spaces] + automata_state_spaces))

        # Check if the brain has the same input space dimensions,
        # but only if the brain has specified an observation space.
        if brain.observation_space and feature_space != brain.observation_space:
            raise ValueError("The brain has incompatible observation space: {} instead of {}"
                             .format(brain.observation_space, feature_space))

        # map every state from the space (N0, N1, ..., Nn) to a discrete space of dimension N0*N1*...*Nn-1
        # self._from_tuple_to_int = TupleFeatureExtractor(feature_space)

    def sync(self, action, state2):
        new_automata_state = tuple(temp_eval.update(action, state2) for temp_eval in self.temporal_evaluators)
        return new_automata_state

    def start(self, state):
        for t in self.temporal_evaluators:
            t.reset()
        automata_states = [temp_eval.get_state() for temp_eval in self.temporal_evaluators]
        transformed_state = self.state_extractor(state, automata_states)
        # return super().start(transformed_state)
        return self.brain.start(transformed_state)

    def state_extractor(self, world_state, automata_states: List):
        # the state is a tuple: (features, A_1 state, ..., A_n state)
        # A_i is the i_th automaton associated to the i_th temporal goal
        state = tuple([self.sensors(world_state)] + list(automata_states))
        return state

    def reward_extractor(self, world_reward, automata_rewards: List):
        res = world_reward + sum(automata_rewards)
        return res

    def get_automata_state(self):
        return tuple(temp_eval.get_state() for temp_eval in self.temporal_evaluators)

    def observe(self, state, action, reward, state2,
                old_automata_state=(), new_automata_state=(), is_terminal_state=False):

        automata_rewards = [te.reward if is_terminal_state and te.is_true() else 0 for te in self.temporal_evaluators]

        # apply reward shaping to the observed automata rewards
        for i in range(len(self.reward_shapers)):
            plus = self.reward_shapers[i](old_automata_state[i], new_automata_state[i], is_terminal_state=is_terminal_state)
            automata_rewards[i] += plus

        old_state  = self.state_extractor(state,  old_automata_state)
        new_state2 = self.state_extractor(state2, new_automata_state)
        new_reward = self.reward_extractor(reward, automata_rewards)

        if old_automata_state!=new_automata_state or is_terminal_state:
            logging.debug("episode: {:6d}, step: {:5d}, old automata states: {:10}, new automata state: {:10}, new reward: {:8.3f}".format(
                self.brain.episode, self.brain.episode_iteration, str(old_automata_state), str(new_automata_state), new_reward
            ))

        return super()._observe(old_state, action, new_reward, new_state2)

    def save(self, filepath):
        super().save(filepath)
        for idx, te in enumerate(self.temporal_evaluators):
            with open(filepath + "/te%02d.dump"%idx, "wb") as fout:
                te.reset()
                pickle.dump(te, fout)

    @staticmethod
    def load(filepath):
        rl_agent = Agent.load(filepath)
        temporal_evaluators_files = sorted([f for f in os.listdir(filepath) if re.match(r'te.*\.pkl', f)])
        temporal_evaluators = []
        for idx, te_name in enumerate(temporal_evaluators_files):
            with open(filepath + "/%s" % te_name, "rb") as fin:
                te = pickle.load(fin)
                temporal_evaluators.append(te)
        temporal_evaluators = temporal_evaluators

        sensors = rl_agent.sensors
        brain = rl_agent.brain

        assert brain and sensors and temporal_evaluators is not None
        return TGAgent(sensors, brain, temporal_evaluators)

    def set_eval(self, eval:bool):
        super().set_eval(eval)
        for te in self.temporal_evaluators:
            te.set_eval(eval)

    def reset(self):
        super().reset()
        for te in self.temporal_evaluators:
            te.reset()



from abc import abstractmethod

import numpy as np
from gym.spaces import Discrete

from rltg.agents.brains.Brain import Brain
from rltg.agents.parameters.Eligibility import Eligibility
from rltg.agents.parameters.Parameter import Constant, AlphaVisitDecay
from rltg.agents.policies.EGreedy import EGreedy
from rltg.agents.policies.Policy import Policy
from rltg.utils.misc import mydefaultdict, AgentObservation


class TDBrain(Brain):
    def __init__(self, observation_space:Discrete, action_space:Discrete, policy:Policy=EGreedy(),
                 gamma=0.99, alpha=None, lambda_=0):
        super().__init__(observation_space, action_space, policy)

        self.gamma = gamma
        self.alpha = Constant(alpha) if alpha is not None else AlphaVisitDecay(action_space)
        self.lambda_ = lambda_

        self._init()

    def _init(self):
        # sparse representation
        self.Q = mydefaultdict(np.zeros((self.action_space.n,)))
        self.eligibility = Eligibility(self.lambda_, self.gamma)

    def start(self, state):
        super().start(state)
        self.eligibility.reset()
        action = self.choose_action(state)
        return action

    def choose_action(self, state, optimal=False):
        Q_values = self.Q[state] # defaultdict, look at __init__
        action = self.policy.choose_action(Q_values, optimal=optimal or self.eval)
        return action

    def step(self, obs:AgentObservation, *args, **kwargs):
        if self.eval:
            return self.choose_action(obs.state2, optimal=self.eval)

        action2 = self.update_Q(obs)
        return action2

    def end(self, obs:AgentObservation, *args, **kwargs):
        super().end(obs)
        if self.eval:
            return
        state, action, reward, state2 = obs.unpack()
        delta = reward - self.Q[state][action]
        for (s, a) in self.eligibility.traces:
            self.Q[s][a] += self.alpha.get(s,a) * delta * self.eligibility.get(s, a)

    @abstractmethod
    def update_Q(self, obs: AgentObservation):
        raise NotImplementedError

    def observe(self, obs:AgentObservation, *args, **kwargs):
        super().observe(obs)
        self.eligibility.to_one(obs.state, obs.action)
        self.alpha.update(obs.state, obs.action)

    def reset(self):
        self._init()


class QLearning(TDBrain):
    def __init__(self, observation_space:Discrete, action_space, policy:Policy=EGreedy(),
                 gamma=0.99, alpha=None, lambda_=0):
        super().__init__(observation_space, action_space, policy, gamma, alpha, lambda_)

    def update_Q(self, obs:AgentObservation):
        state, action, reward, state2 = obs.unpack()

        # Q-Learning
        action2 = self.choose_action(state2)
        Qa = np.max(self.Q[state2])
        actions_star = np.argwhere(self.Q[state2] == Qa).flatten().tolist()

        delta = reward + self.gamma * Qa - self.Q[state][action]
        for (s, a) in set(self.eligibility.traces.keys()):
            self.Q[s][a] += self.alpha.get(s,a) * delta * self.eligibility.get(s, a)
            # Q-Learning
            if action2 in actions_star:
                self.eligibility.update(s, a)
            else:
                self.eligibility.to_zero(s, a)

        return action2


class Sarsa(TDBrain):
    def __init__(self, observation_space:Discrete, action_space, policy:Policy=EGreedy(),
                 gamma=0.99, alpha=None, lambda_=0.0):
        super().__init__(observation_space, action_space, policy, gamma, alpha, lambda_)


    def update_Q(self, obs:AgentObservation):
        state, action, reward, state2 = obs.unpack()

        # SARSA
        action2 = self.choose_action(state2)
        Qa = self.Q[state2][action2]

        delta = reward + self.gamma * Qa - self.Q[state][action]
        # if delta!=0.0:
        #     print(len(set(self.eligibility.traces.keys())), delta, reward, Qa, self.Q[state][action])
        for (s, a) in set(self.eligibility.traces.keys()):
            self.Q[s][a] += self.alpha.get(s,a) * delta * self.eligibility.get(s, a)
            # SARSA
            self.eligibility.update(s, a)

        return action2



import pickle

import numpy as np
import matplotlib.pyplot as plt

class StatsManager(object):

    def __init__(self, name="stats", window_size=100):
        self.name = name
        self.window_size = window_size
        self._init_arrays()

    def _init_arrays(self):
        self.steps_taken = np.array([], dtype=np.int32)
        self.total_reward_history = np.array([], dtype=np.int32)
        self.avg_reward_history = np.array([], dtype=np.float32)
        self.std_reward_history = np.array([], dtype=np.float32)
        self.explored_states_history = np.array([], dtype=np.int32)
        self.goals = np.array([], dtype=np.bool)

    def update(self, steps, n_states, total_reward, goal):
        self.steps_taken = np.append(self.steps_taken, steps)
        self.explored_states_history = np.append(self.explored_states_history, n_states)
        self.total_reward_history = np.append(self.total_reward_history, total_reward)
        avg_reward = np.mean(self.total_reward_history[-self.window_size:])
        std_reward = np.std(self.total_reward_history[-self.window_size:])
        self.avg_reward_history = np.append(self.avg_reward_history, avg_reward)
        self.std_reward_history = np.append(self.std_reward_history, std_reward)

        self.goals = np.append(self.goals, goal)


    def print_summary(self, episode_number, step, n_states, total_reward, epsilon, goal):
        w = self.window_size
        return "Episode: {:6d}, Step: {:5d}, Explored States: {:7d}, Total Reward: {:8.2f}, AvgReward: {:8.2f}, StdReward: {:8.2f}, Epsilon: {:05.4f}, Goal: {:5}, GoalPerc: {:5.2f}"\
              .format(episode_number, step, n_states, total_reward,
                      np.mean(self.total_reward_history[-w:]),
                      np.std(self.total_reward_history[-w:]),
                      epsilon, str(goal), np.mean(self.goals[-w:])*100)

    def plot(self):
        plt.figure(1)
        plt.title("Total reward history")
        plt.plot(self.total_reward_history)
        plt.figure(2)
        plt.title("Average total reward history")
        plt.plot(list(self.avg_reward_history))
        plt.figure(3)
        plt.title("Standard deviation total reward history")
        plt.plot(list(self.std_reward_history))
        plt.show()

    def to_csv(self, filepath):
        with open(filepath + ".csv", "w") as f:
            f.write("Episode;Steps;Total reward;Explored states;Goal\n")
            for ep, (steps, total_rewards, explored_states, goals) in enumerate(zip(
                self.steps_taken, self.total_reward_history, self.explored_states_history, self.goals
            )):
                f.write(";".join(map(str,[ep, steps, total_rewards, explored_states, goals])) + "\n")

    def reset(self):
        self._init_arrays()



import logging
from copy import copy
from itertools import chain
from typing import Any, Set, List

from flloat.base.Alphabet import Alphabet
from flloat.base.Symbol import Symbol
from flloat.flloat import DFAOTF
from flloat.semantics.pl import PLInterpretation
from flloat.syntax.ldlf import LDLfFormula
from pythomata.base.DFA import DFA

from rltg.logic.RewardAutomaton import RewardAutomaton
from rltg.logic.RewardAutomatonSimulator import RewardSimulator
from rltg.utils.misc import compute_levels, _potential_function, mydefaultdict


class PartialRewardAutomaton(RewardAutomaton, RewardSimulator):
    def __init__(self, alphabet:Alphabet, f:LDLfFormula, reward, gamma=0.99):

        self.dfaotf = DFAOTF(f)
        self.alphabet = alphabet
        self.f = f
        self.reward = reward

        self.state2id = {self.dfaotf.cur_state: 0}
        self.id2state = {0: self.dfaotf.cur_state}

        self.states = {0}
        self.initial_state = 0
        self.accepting_states = {0} if self.dfaotf.is_true() else set()

        self.transition_function = {}
        self.inverse_transition_function = {}
        self.links = mydefaultdict(set())

        self.failure_states = set() if self._is_failed_state(self.dfaotf.cur_state) else {0}

        self.gamma = gamma
        self.episode = 0
        self.it = 0

        self.potentials = mydefaultdict(0)
        self.potentials_prime = mydefaultdict(0)


    def dfs(self, dfa, state, visited, border_states):
        leaves = []
        visited.add(state)
        exists_next = False
        if state in self.accepting_states:
            return leaves
        for a in dfa.transition_function.get(state, {}):
            next_state = dfa.transition_function[state][a]
            if next_state not in visited and not self._is_failed_state(self.id2state[next_state]):
                exists_next = True
                leaves += self.dfs(dfa, next_state, visited, border_states)
        if not exists_next and not self._is_failed_state(self.id2state[state]) and state in border_states:
            leaves.append(state)
        return leaves


    def _compute_levels(self):
        temp_dfa = DFA(self.alphabet, self.states, self.initial_state, self.accepting_states, self.transition_function)
        state2level, max_level, failure_states = compute_levels(temp_dfa, temp_dfa.accepting_states)

        border_states = set(filter(lambda x: not self._is_failed_state(self.id2state[x]), failure_states))
        true_failure_state = set(filter(lambda x: self._is_failed_state(self.id2state[x]), failure_states))
        true_states = set(filter(lambda x: not self._is_failed_state(self.id2state[x]) and x not in border_states, temp_dfa.states))
        potentials = {s: _potential_function(s, self.initial_state, state2level, self.reward) for s in chain(true_failure_state, true_states)}

        # compute levels from initial state
        # level = 0
        # visited = set()
        # to_visit = set()
        # to_visit_next = {self.initial_state}
        # reachability_levels = {self.initial_state:0}
        #
        # changed = True
        # while changed:
        #     changed = False
        #     level += 1
        #     to_visit = to_visit_next
        #     to_visit_next = set()
        #
        #     for s in to_visit:
        #         if s in visited:
        #             continue
        #         else:
        #             visited.add(s)
        #         s_potential = potentials[s]
        #         for a in temp_dfa.transition_function.get(s, []):
        #             s_prime = temp_dfa.transition_function[s][a]
        #
        #             if s_prime not in reachability_levels:
        #                 reachability_levels[s_prime] = level
        #             elif level < reachability_levels[s_prime]:
        #                 reachability_levels[s_prime] = level
        #
        #             old_potential = potentials.get(s_prime, self.reward)
        #             # if potential is computed correctly
        #             if s_prime not in failure_states:
        #                 potentials[s_prime] = _potential_function(s, self.initial_state, state2level, self.reward)
        #             elif s_potential < old_potential:
        #                 potentials[s_prime] = s_potential
        #
        #             to_visit_next.add(s_prime)
        #             changed = True


        # least fixpoint with the new state just discovered
        leaves = self.dfs(temp_dfa, self.initial_state, set(), border_states)
        border_state2level, border_max_level, _ = compute_levels(temp_dfa, set(leaves))
        for b in border_states:
            potentials[b] = _potential_function(b, self.initial_state, border_state2level, self.reward)

        logging.debug("border states: {:15}, leaves: {:10}".format(str(border_states), str(leaves)))

        return potentials



    def potential_function(self, q, is_terminal_state=False, is_prime=False):
        if is_terminal_state:
            return 0
        elif is_prime:
            return self.potentials_prime[q]
        else:
            return self.potentials[q]


    def _update_from_transition(self, old_state, label, new_state):

        old_state_id, old_state_changed = self._add_state(old_state)
        new_state_id, new_state_changed = self._add_state(new_state)
        transition_already_exists = False
        is_failure_state = False
        is_final_state = False

        if old_state_id not in self.transition_function or label not in self.transition_function[old_state_id]:
            transition_already_exists = new_state_id in self.links[old_state_id]
            self.links[old_state_id].add(new_state_id)
            self.transition_function.setdefault(old_state_id, {})[label] = new_state_id

        if self._is_failed_state(new_state) and new_state_id not in self.failure_states:
            self.failure_states.add(new_state_id)
            is_failure_state = True

        elif self.dfaotf._is_true(new_state) and new_state_id not in self.accepting_states:
            self.accepting_states.add(new_state_id)
            is_final_state = True

        changed = old_state_changed or (new_state_changed and not is_failure_state and not transition_already_exists)

        if changed:

            self.potentials = self.potentials_prime
            self.potentials_prime = self._compute_levels()

            dfa = DFA(self.alphabet, frozenset(self.states), self.initial_state, frozenset(self.accepting_states), self.transition_function)
            dfa.to_dot("automata/%s/%s-%s" % (str(self.dfaotf.f)[:9],self.episode, self.it))

            logging.debug(
                "episode: {:6d}, step: {:5d}, update automaton on-the-fly, prev_state: {:3}, next_state_id: {:3}, label: {:10}, is fail: {:5}, is final: {:5}".format(
                    self.episode, self.it, old_state_id, new_state_id, str(label), str(new_state_id in self.failure_states), str(new_state_id in self.accepting_states)
                ))
            logging.debug("new potentials: %s", self.potentials_prime)
        else:
            self.potentials = copy(self.potentials_prime)

    def _add_state(self, new_state):
        changed = False
        new_state_id = self.state2id.get(new_state, None)
        if new_state_id is None:
            # first time we see the state
            changed = True
            new_state_id = len(self.states)
            self.states.add(new_state_id)
            self.id2state[new_state_id] = new_state
            self.state2id[new_state] = new_state_id

        return new_state_id, changed

    def _is_failed_state(self, state):
        return state == frozenset()

    def is_failed(self):
        return self._is_failed_state(self.dfaotf.cur_state)

    def make_transition(self, s:Set[Symbol], **kwargs) -> Any:
        old_state = self.dfaotf.get_current_state()
        i = PLInterpretation(s)
        self.dfaotf.make_transition(i)
        new_state = self.dfaotf.cur_state
        self._update_from_transition(old_state, i, new_state)
        self.it += 1
        return self.get_current_state()

    def is_true(self) -> bool:
        return self.dfaotf.is_true()

    def reset(self) -> Any:
        self.dfaotf.reset()
        self.episode += 1
        self.it = 0

    def get_current_state(self):
        return self.state2id[self.dfaotf.get_current_state()]

    def word_acceptance(self, word: List[Symbol]):
        raise NotImplementedError

